@@ -25,7 +25,7 @@ import copy
 import warnings
 from functools import reduce
 
-import numpy
+import numpy as np
 
 from ..util import (
     IS_PY3,
@@ -152,7 +152,7 @@ class TableHDU(HDUBase):
         else:
             return False
 
-    def where(self, expression):
+    def where(self, expression, firstrow=None, lastrow=None):
         """
         Return the indices where the expression evaluates to true.
 
@@ -161,8 +161,22 @@ class TableHDU(HDUBase):
         expression: string
             A fits row selection expression.  E.g.
             "x > 3 && y < 5"
+        firstrow, lastrow : int
+            Range of rows for evaluation.  This follows the Python list
+            slice convention that the last element is not included.
         """
-        return self._FITS.where(self._ext+1, expression)
+        if firstrow is None:
+            firstrow = 0
+        elif firstrow < 0:
+            raise ValueError('firstrow cannot be negative')
+        if lastrow is None:
+            lastrow = self._info['nrows']
+        elif lastrow < firstrow:
+            raise ValueError('lastrow cannot be less than firstrow')
+        elif lastrow > self._info['nrows']:
+            raise ValueError('lastrow cannot be greater than nrows')
+        nrows = lastrow - firstrow
+        return self._FITS.where(self._ext+1, expression, firstrow+1, nrows)
 
     def write(self, data, firstrow=0, columns=None, names=None, slow=False,
               **keys):
@@ -220,7 +234,7 @@ class TableHDU(HDUBase):
             colnums_all = [self._extract_colnum(c) for c in columns_all]
             names = [self.get_colname(c) for c in colnums_all]
 
-            isobj = numpy.zeros(len(data_list), dtype=bool)
+            isobj = np.zeros(len(data_list), dtype=bool)
             for i in xrange(len(data_list)):
                 isobj[i] = is_object(data_list[i])
 
@@ -321,7 +335,7 @@ class TableHDU(HDUBase):
 
         if not data.flags['C_CONTIGUOUS']:
             # this always makes a copy
-            data_send = numpy.ascontiguousarray(data)
+            data_send = np.ascontiguousarray(data)
             # this is a copy, we can make sure it is native
             # and modify in place if needed
             array_to_native(data_send, inplace=True)
@@ -381,17 +395,6 @@ class TableHDU(HDUBase):
         if col_shape is not None and not isinstance(col_shape, tuple):
             col_shape = (col_shape,)
 
-        """
-        print('column name:',col_name)
-        print(data.shape)
-        print('col tdim', info['tdim'])
-        print('column dtype:',npy_type)
-        print('input dtype:',this_npy_type)
-        print('column shape:',col_shape)
-        print('input shape:',this_shape)
-        print()
-        """
-
         # this mismatch is OK
         if npy_type == 'i1' and this_npy_type == 'b1':
             this_npy_type = 'i1'
@@ -490,7 +493,7 @@ class TableHDU(HDUBase):
             # the numpy API is probably safer
             # this also avoids doing a dtype conversion on every array
             # element which could b expensive
-            descr = numpy.empty(1).astype(data.dtype).astype('S').dtype.descr
+            descr = np.empty(1).astype(data.dtype).astype('S').dtype.descr
         else:
             descr = data.dtype.descr
 
@@ -581,7 +584,7 @@ class TableHDU(HDUBase):
         if isinstance(rows, slice):
             rows = self._process_slice(rows)
             if rows.step is not None and rows.step != 1:
-                rows = numpy.arange(
+                rows = np.arange(
                     rows.start+1,
                     rows.stop+1,
                     rows.step,
@@ -590,7 +593,7 @@ class TableHDU(HDUBase):
                 # rows must be 1-offset
                 rows = slice(rows.start+1, rows.stop+1)
         else:
-            rows = self._extract_rows(rows)
+            rows, sortind = self._extract_rows(rows, sort=True)
             # rows must be 1-offset
             rows += 1
 
@@ -751,7 +754,7 @@ class TableHDU(HDUBase):
         dtype, offsets, isvar = self.get_rec_dtype(
             colnums=colnums, vstorage=vstorage)
 
-        w, = numpy.where(isvar == True)  # noqa
+        w, = np.where(isvar == True)  # noqa
         has_tbit = self._check_tbit()
 
         if w.size > 0:
@@ -761,8 +764,11 @@ class TableHDU(HDUBase):
                 _vstorage = vstorage
             colnums = self._extract_colnums()
             rows = None
-            array = self._read_rec_with_var(colnums, rows, dtype,
-                                            offsets, isvar, _vstorage)
+            sortind = None
+            array = self._read_rec_with_var(
+                colnums, rows, sortind, dtype,
+                offsets, isvar, _vstorage,
+            )
         elif has_tbit:
             # drop down to read_columns since we can't stuff into a
             # contiguous array
@@ -774,7 +780,7 @@ class TableHDU(HDUBase):
         else:
             firstrow = 1  # noqa - not used?
             nrows = self._info['nrows']
-            array = numpy.zeros(nrows, dtype=dtype)
+            array = np.zeros(nrows, dtype=dtype)
 
             self._FITS.read_as_rec(self._ext+1, 1, nrows, array)
 
@@ -885,10 +891,10 @@ class TableHDU(HDUBase):
                 rows=rows, vstorage=vstorage,
                 upper=upper, lower=lower, trim_strings=trim_strings)
 
-        rows = self._extract_rows(rows)
+        rows, sortind = self._extract_rows(rows)
         dtype, offsets, isvar = self.get_rec_dtype(vstorage=vstorage)
 
-        w, = numpy.where(isvar == True)  # noqa
+        w, = np.where(isvar == True)  # noqa
         if w.size > 0:
             if vstorage is None:
                 _vstorage = self._vstorage
@@ -896,10 +902,11 @@ class TableHDU(HDUBase):
                 _vstorage = vstorage
             colnums = self._extract_colnums()
             return self._read_rec_with_var(
-                colnums, rows, dtype, offsets, isvar, _vstorage)
+                colnums, rows, sortind, dtype, offsets, isvar, _vstorage,
+            )
         else:
-            array = numpy.zeros(rows.size, dtype=dtype)
-            self._FITS.read_rows_as_rec(self._ext+1, array, rows)
+            array = np.zeros(rows.size, dtype=dtype)
+            self._FITS.read_rows_as_rec(self._ext+1, array, rows, sortind)
 
             array = self._maybe_decode_fits_ascii_strings_to_unicode_py3(array)
 
@@ -970,35 +977,40 @@ class TableHDU(HDUBase):
             return self.read_column(
                 columns,
                 rows=rows, vstorage=vstorage,
-                upper=upper, lower=lower, trim_strings=trim_strings)
+                upper=upper, lower=lower, trim_strings=trim_strings,
+            )
 
         # if rows is None still returns None, and is correctly interpreted
         # by the reader to mean all
-        rows = self._extract_rows(rows)
+        rows, sortind = self._extract_rows(rows)
 
         # this is the full dtype for all columns
         dtype, offsets, isvar = self.get_rec_dtype(
             colnums=colnums, vstorage=vstorage)
 
-        w, = numpy.where(isvar == True)  # noqa
+        w, = np.where(isvar == True)  # noqa
         if w.size > 0:
             if vstorage is None:
                 _vstorage = self._vstorage
             else:
                 _vstorage = vstorage
             array = self._read_rec_with_var(
-                colnums, rows, dtype, offsets, isvar, _vstorage)
+                colnums, rows, sortind, dtype, offsets, isvar, _vstorage,
+            )
         else:
 
             if rows is None:
                 nrows = self._info['nrows']
             else:
                 nrows = rows.size
-            array = numpy.zeros(nrows, dtype=dtype)
+
+            array = np.zeros(nrows, dtype=dtype)
 
             colnumsp = colnums[:].copy()
             colnumsp[:] += 1
-            self._FITS.read_columns_as_rec(self._ext+1, colnumsp, array, rows)
+            self._FITS.read_columns_as_rec(
+                self._ext+1, colnumsp, array, rows, sortind
+            )
 
             array = self._maybe_decode_fits_ascii_strings_to_unicode_py3(array)
 
@@ -1065,7 +1077,7 @@ class TableHDU(HDUBase):
                 DeprecationWarning, stacklevel=2)
 
         if self._info['hdutype'] == ASCII_TBL:
-            rows = numpy.arange(firstrow, lastrow, step, dtype='i8')
+            rows = np.arange(firstrow, lastrow, step, dtype='i8')
             return self.read_ascii(
                 rows=rows, vstorage=vstorage,
                 upper=upper, lower=lower, trim_strings=trim_strings)
@@ -1080,24 +1092,25 @@ class TableHDU(HDUBase):
 
         dtype, offsets, isvar = self.get_rec_dtype(vstorage=vstorage)
 
-        w, = numpy.where(isvar == True)  # noqa
+        w, = np.where(isvar == True)  # noqa
         if w.size > 0:
             if vstorage is None:
                 _vstorage = self._vstorage
             else:
                 _vstorage = vstorage
-            rows = numpy.arange(firstrow, lastrow, step, dtype='i8')
+            rows = np.arange(firstrow, lastrow, step, dtype='i8')
+            sortind = np.arange(rows.size, dtype='i8')
             colnums = self._extract_colnums()
             array = self._read_rec_with_var(
-                colnums, rows, dtype, offsets, isvar, _vstorage)
+                colnums, rows, sortind, dtype, offsets, isvar, _vstorage)
         else:
             if step != 1:
-                rows = numpy.arange(firstrow, lastrow, step, dtype='i8')
+                rows = np.arange(firstrow, lastrow, step, dtype='i8')
                 array = self.read(rows=rows)
             else:
                 # no +1 because lastrow is non-inclusive
                 nrows = lastrow - firstrow
-                array = numpy.zeros(nrows, dtype=dtype)
+                array = np.zeros(nrows, dtype=dtype)
 
                 # only first needs to be +1.  This is becuase the c code is
                 # inclusive
@@ -1149,14 +1162,14 @@ class TableHDU(HDUBase):
             colnums = self._extract_colnums()
 
         descr = []
-        isvararray = numpy.zeros(len(colnums), dtype=bool)
+        isvararray = np.zeros(len(colnums), dtype=bool)
         for i, colnum in enumerate(colnums):
             dt, isvar = self.get_rec_column_descr(colnum, _vstorage)
             descr.append(dt)
             isvararray[i] = isvar
-        dtype = numpy.dtype(descr)
+        dtype = np.dtype(descr)
 
-        offsets = numpy.zeros(len(colnums), dtype='i8')
+        offsets = np.zeros(len(colnums), dtype='i8')
         for i, n in enumerate(dtype.names):
             offsets[i] = dtype.fields[n][1]
         return dtype, offsets, isvararray
@@ -1306,7 +1319,7 @@ class TableHDU(HDUBase):
         return descr, isvar
 
     def _read_rec_with_var(
-            self, colnums, rows, dtype, offsets, isvar, vstorage):
+            self, colnums, rows, sortind, dtype, offsets, isvar, vstorage):
         """
         Read columns from a table into a rec array, including variable length
         columns.  This is special because, for efficiency, it involves reading
@@ -1322,19 +1335,22 @@ class TableHDU(HDUBase):
             nrows = self._info['nrows']
         else:
             nrows = rows.size
-        array = numpy.zeros(nrows, dtype=dtype)
+        array = np.zeros(nrows, dtype=dtype)
 
         # read from the main table first
-        wnotvar, = numpy.where(isvar == False)  # noqa
+        wnotvar, = np.where(isvar == False)  # noqa
         if wnotvar.size > 0:
             # this will be contiguous (not true for slices)
             thesecol = colnumsp[wnotvar]
             theseoff = offsets[wnotvar]
-            self._FITS.read_columns_as_rec_byoffset(self._ext+1,
-                                                    thesecol,
-                                                    theseoff,
-                                                    array,
-                                                    rows)
+            self._FITS.read_columns_as_rec_byoffset(
+                self._ext+1,
+                thesecol,
+                theseoff,
+                array,
+                rows,
+                sortind,
+            )
             for i in xrange(thesecol.size):
 
                 name = array.dtype.names[wnotvar[i]]
@@ -1343,13 +1359,14 @@ class TableHDU(HDUBase):
                     array,
                     name,
                     self._info['colinfo'][colnum]['tscale'],
-                    self._info['colinfo'][colnum]['tzero'])
+                    self._info['colinfo'][colnum]['tzero'],
+                )
 
         array = self._maybe_decode_fits_ascii_strings_to_unicode_py3(array)
 
         # now read the variable length arrays we may be able to speed this up
         # by storing directly instead of reading first into a list
-        wvar, = numpy.where(isvar == True)  # noqa
+        wvar, = np.where(isvar == True)  # noqa
         if wvar.size > 0:
             # this will be contiguous (not true for slices)
             thesecol = colnumsp[wvar]
@@ -1357,7 +1374,8 @@ class TableHDU(HDUBase):
                 colnump = thesecol[i]
                 name = array.dtype.names[wvar[i]]
                 dlist = self._FITS.read_var_column_as_list(
-                    self._ext+1, colnump, rows)
+                    self._ext+1, colnump, rows, sortind,
+                )
 
                 if (isinstance(dlist[0], str) or
                         (IS_PY3 and isinstance(dlist[0], bytes))):
@@ -1369,11 +1387,15 @@ class TableHDU(HDUBase):
                     # storing in object array
                     # get references to each, no copy made
                     for irow, item in enumerate(dlist):
+                        if sortind is not None:
+                            irow = sortind[irow]
                         if IS_PY3 and isinstance(item, bytes):
                             item = item.decode('ascii')
                         array[name][irow] = item
                 else:
                     for irow, item in enumerate(dlist):
+                        if sortind is not None:
+                            irow = sortind[irow]
                         if IS_PY3 and isinstance(item, bytes):
                             item = item.decode('ascii')
 
@@ -1393,19 +1415,30 @@ class TableHDU(HDUBase):
 
         return array
 
-    def _extract_rows(self, rows):
+    def _extract_rows(self, rows, sort=False):
         """
         Extract an array of rows from an input scalar or sequence
         """
         if rows is not None:
-            rows = numpy.array(rows, ndmin=1, copy=False, dtype='i8')
-            # returns unique, sorted
-            rows = numpy.unique(rows)
+            rows = np.array(rows, ndmin=1, copy=False, dtype='i8')
+            if sort:
+                rows = np.unique(rows)
+                return rows, None
+
+            # returns unique, sorted.  Force i8 for 32-bit systems
+            sortind = np.array(rows.argsort(), dtype='i8', copy=False)
 
             maxrow = self._info['nrows']-1
-            if len(rows) > 0 and (rows[0] < 0 or rows[-1] > maxrow):
-                raise ValueError("rows must be in [%d,%d]" % (0, maxrow))
-        return rows
+            if rows.size > 0:
+                firstrow = rows[sortind[0]]
+                lastrow = rows[sortind[-1]]
+
+                if len(rows) > 0 and (firstrow < 0 or lastrow > maxrow):
+                    raise ValueError("rows must be in [%d,%d]" % (0, maxrow))
+        else:
+            sortind = None
+
+        return rows, sortind
 
     def _process_slice(self, arg):
         """
@@ -1459,7 +1492,7 @@ class TableHDU(HDUBase):
             return None
         if stop < start:
             raise ValueError("start is greater than stop in slice")
-        return numpy.arange(tstart, tstop, step, dtype='i8')
+        return np.arange(tstart, tstop, step, dtype='i8')
 
     def _fix_range(self, num, isslice=True):
         """
@@ -1510,10 +1543,10 @@ class TableHDU(HDUBase):
         Scale the input array
         """
         if scale != 1.0:
-            sval = numpy.array(scale, dtype=array.dtype)
+            sval = np.array(scale, dtype=array.dtype)
             array *= sval
         if zero != 0.0:
-            zval = numpy.array(zero, dtype=array.dtype)
+            zval = np.array(zero, dtype=array.dtype)
             array += zval
 
     def _maybe_trim_strings(self, array, trim_strings=False, **keys):
@@ -1559,7 +1592,7 @@ class TableHDU(HDUBase):
         If input is a fits bool, convert to numpy boolean
         """
 
-        output = (array.view(numpy.int8) == ord('T')).astype(bool)
+        output = (array.view(np.int8) == ord('T')).astype(bool)
         return output
 
     def _get_tbl_numpy_dtype(self, colnum, include_endianness=True):
@@ -1620,7 +1653,7 @@ class TableHDU(HDUBase):
 
         flags = set()
         #
-        if isinstance(arg, (tuple, list, numpy.ndarray)):
+        if isinstance(arg, (tuple, list, np.ndarray)):
             # a sequence was entered
             if isstring(arg[0]):
                 result = arg
@@ -1643,12 +1676,12 @@ class TableHDU(HDUBase):
             # Probably should apply some more checking on this
             result = arg
             flags.add('isrows')
-            if numpy.ndim(arg) == 0:
+            if np.ndim(arg) == 0:
                 flags.add('isscalar')
 
         return result, flags
 
-    def _read_var_column(self, colnum, rows, vstorage):
+    def _read_var_column(self, colnum, rows, sortind, vstorage):
         """
 
         first read as a list of arrays, then copy into either a fixed length
@@ -1661,7 +1694,9 @@ class TableHDU(HDUBase):
         else:
             stype = str
 
-        dlist = self._FITS.read_var_column_as_list(self._ext+1, colnum+1, rows)
+        dlist = self._FITS.read_var_column_as_list(
+            self._ext+1, colnum+1, rows, sortind,
+        )
 
         if vstorage == 'fixed':
             tform = self._info['colinfo'][colnum]['tform']
@@ -1684,19 +1719,24 @@ class TableHDU(HDUBase):
 
             if isinstance(dlist[0], stype):
                 descr = 'S%d' % max_size
-                array = numpy.fromiter(dlist, descr)
+                array = np.fromiter(dlist, descr)
                 if IS_PY3:
                     array = array.astype('U', copy=False)
             else:
                 descr = dlist[0].dtype.str
-                array = numpy.zeros((len(dlist), max_size), dtype=descr)
+                array = np.zeros((len(dlist), max_size), dtype=descr)
 
                 for irow, item in enumerate(dlist):
+                    if sortind is not None:
+                        irow = sortind[irow]
                     ncopy = len(item)
                     array[irow, 0:ncopy] = item[:]
         else:
-            array = numpy.zeros(len(dlist), dtype='O')
+            array = np.zeros(len(dlist), dtype='O')
             for irow, item in enumerate(dlist):
+                if sortind is not None:
+                    irow = sortind[irow]
+
                 if IS_PY3 and isinstance(item, bytes):
                     item = item.decode('ascii')
                 array[irow] = item
@@ -1708,18 +1748,18 @@ class TableHDU(HDUBase):
         Extract an array of columns from the input
         """
         if columns is None:
-            return numpy.arange(self._ncol, dtype='i8')
+            return np.arange(self._ncol, dtype='i8')
 
-        if not isinstance(columns, (tuple, list, numpy.ndarray)):
+        if not isinstance(columns, (tuple, list, np.ndarray)):
             # is a scalar
             return self._extract_colnum(columns)
 
-        colnums = numpy.zeros(len(columns), dtype='i8')
+        colnums = np.zeros(len(columns), dtype='i8')
         for i in xrange(colnums.size):
             colnums[i] = self._extract_colnum(columns[i])
 
         # returns unique sorted
-        colnums = numpy.unique(colnums)
+        colnums = np.unique(colnums)
         return colnums
 
     def _extract_colnum(self, col):
@@ -1969,7 +2009,7 @@ class AsciiTableHDU(TableHDU):
                 columns, rows=rows, vstorage=vstorage,
                 upper=upper, lower=lower, trim_strings=trim_strings)
 
-        rows = self._extract_rows(rows)
+        rows, sortind = self._extract_rows(rows)
         if rows is None:
             nrows = self._info['nrows']
         else:
@@ -1977,33 +2017,34 @@ class AsciiTableHDU(TableHDU):
 
         # if rows is None still returns None, and is correctly interpreted
         # by the reader to mean all
-        rows = self._extract_rows(rows)
+        rows, sortind = self._extract_rows(rows)
 
         # this is the full dtype for all columns
         dtype, offsets, isvar = self.get_rec_dtype(
             colnums=colnums, vstorage=vstorage)
-        array = numpy.zeros(nrows, dtype=dtype)
+        array = np.zeros(nrows, dtype=dtype)
 
         # note reading into existing data
-        wnotvar, = numpy.where(isvar == False)  # noqa
+        wnotvar, = np.where(isvar == False)  # noqa
         if wnotvar.size > 0:
             for i in wnotvar:
                 colnum = colnums[i]
                 name = array.dtype.names[i]
                 a = array[name].copy()
-                self._FITS.read_column(self._ext+1, colnum+1, a, rows)
+                self._FITS.read_column(self._ext+1, colnum+1, a, rows, sortind)
                 array[name] = a
                 del a
 
         array = self._maybe_decode_fits_ascii_strings_to_unicode_py3(array)
 
-        wvar, = numpy.where(isvar == True)  # noqa
+        wvar, = np.where(isvar == True)  # noqa
         if wvar.size > 0:
             for i in wvar:
                 colnum = colnums[i]
                 name = array.dtype.names[i]
                 dlist = self._FITS.read_var_column_as_list(
-                    self._ext+1, colnum+1, rows)
+                    self._ext+1, colnum+1, rows, sortind,
+                )
                 if (isinstance(dlist[0], str) or
                         (IS_PY3 and isinstance(dlist[0], bytes))):
                     is_string = True
@@ -2014,11 +2055,15 @@ class AsciiTableHDU(TableHDU):
                     # storing in object array
                     # get references to each, no copy made
                     for irow, item in enumerate(dlist):
+                        if sortind is not None:
+                            irow = sortind[irow]
                         if IS_PY3 and isinstance(item, bytes):
                             item = item.decode('ascii')
                         array[name][irow] = item
                 else:
                     for irow, item in enumerate(dlist):
+                        if sortind is not None:
+                            irow = sortind[irow]
                         if IS_PY3 and isinstance(item, bytes):
                             item = item.decode('ascii')
                         if is_string:
@@ -2252,10 +2297,10 @@ def _trim_strings(data):
         # run through each field separately
         for n in names:
             if data[n].dtype.descr[0][1][1] in ['S', 'U']:
-                data[n] = numpy.char.rstrip(data[n])
+                data[n] = np.char.rstrip(data[n])
     else:
         if data.dtype.descr[0][1][1] in ['S', 'U']:
-            data[:] = numpy.char.rstrip(data[:])
+            data[:] = np.char.rstrip(data[:])
 
 
 def _extract_vararray_max(tform):
